#!/usr/bin/env python

__version__ = "0.5.1"
import os
import shutil
import logging
from pathlib import Path
import libgwak.manifest
import libgwak.zy


def _filter_dir(path: str) -> Path:
    res = Path(path).resolve()
    if not res.is_dir():
        raise NotADirectoryError(path)
    return res


class ZYFormatter(logging.Formatter):
    def formatTime(self, *args, **kwargs) -> str:
        return libgwak.zy.zy.encode()


class Gwak:
    MANIFEST: str = 'gwak'
    GWAK: str = '._gwak'
    FILTERGLOB: str = '[!.]*'

    minsize: int = 256
    mindupe: int = 2

    def __init__(self, params, **kwargs):
        self._params = params
        self._params.__dict__.update(kwargs)
        self._logger = self._params.logger
        if hasattr(self._params, 'minsize'):
            self.minsize = self._params.minsize
        if hasattr(self._params, 'mindupe'):
            self.minsize = self._params.mindupe

    def _rmdir(self, path: Path) -> bool:
        if not path.is_dir() or any(path.iterdir()):
            return False
        self._logger.debug(f"rmdir [{path}]")
        if not self._params.dry_run:
            path.rmdir()
        return True

    def _is_smol(self, size: str) -> bool:
        return int(size, 16) < self.minsize

    def _bury(self, size: str, hash: str, file: Path) -> dict:
        body_dir = self._params.grave / size
        if not self._params.dry_run:
            body_dir.mkdir(parents = True, exist_ok = True)
        body = body_dir / hash
        link = body if self._params.isabs else os.path.relpath(body, file.parent)
        self._logger.debug(f"symlink [{link}]")
        if not self._params.dry_run:
            if body.is_file():
                file.unlink()
            else:
                file.rename(body)
            file.symlink_to(link)
        return {
            'link': str(link),
            'body': str(file)
        }

    def _exhume(self, file: Path, links: list) -> bool:
        self._logger.info(f"exhuming [{file}]")
        for link in links:
            self._logger.debug(f"copyfile [{link}]")
            if not self._params.dry_run:
                link.unlink()
                shutil.copyfile(file, link)
        if self._params.force and not self._params.dry_run:
            self._logger.debug(f"remove [{file}]")
            file.unlink()
        return True

    def _dedupe(self, gwaks: dict):
        for size, gwak in gwaks.items():
            if self._is_smol(size) and not self._params.force:
                self._logger.debug(f"skipping small files [{size}]")
                continue
            for hash, files in gwak.items():
                if len(files) < self.mindupe and not self._params.force:
                    self._logger.debug(f"skipping unique files [{hash}]")
                    continue
                for file in files:
                    yield self._bury(size, hash, file)

    def dedupe(self, gwaks: dict) -> list:
        return list(self._dedupe(gwaks))

    def _redupe(self, gwaks: dict):
        for size, gwak in gwaks.items():
            sizedir = self._params.grave / size
            for hash, links in gwak.items():
                hashdir = sizedir / hash
                file = sizedir / hash
                if not file.is_file():
                    self._logger.debug(f"skipping missing file [{file}]")
                    continue
                yield self._exhume(file, links)
                if self._params.force:
                    self._rmdir(hashdir)
            if self._params.force:
                self._rmdir(sizedir)

    def redupe(self, gwaks: dict) -> bool:
        return all(self._redupe(gwaks))

    def _validate_body(self, file: Path, size: str, hash: str) -> bool:
        body = file.read_bytes()
        if size != libgwak.manifest.gwak_size(body):
            self._logger.warning(f"size mismatch [{file}]")
            return False
        if hash != libgwak.manifest.gwak_hash(body):
            self._logger.warning(f"hash mismatch [{file}]")
            return False
        return True

    def _validate_files(self, gwaks: dict):
        for size, gwak in gwaks.items():
            for hash, files in gwak.items():
                for file in files:
                    if not file.is_file():
                        self._logger.warning("no such file [{file}]")
                        continue
                    yield self._validate_body(file, size, hash)

    def validate_files(self, gwaks: dict) -> bool:
        return all(self._validate_files(gwaks))

    def _validate_grave(self, gwaks: dict):
        for size, gwak in gwaks.items():
            for hash in gwak:
                file = self._params.grave / size / hash
                if not file.is_file():
                    self._logger.info(f"no such body [{file}]")
                    continue
                yield self._validate_body(file, size, hash)

    def validate_grave(self, gwaks: dict) -> bool:
        return all(self._validate_grave(gwaks))


def main():
    import sys
    import argparse
    import json
    __LOGFORMAT = '%(asctime)s %(levelname)-8s | %(message)s'
    def run_gwak():
        global __params
        parser = argparse.ArgumentParser(description = "Gwak a directory by burying filebodies and replacing them with symlinks.")
        actions = parser.add_mutually_exclusive_group()
        parser.add_argument('-V', '--version', action = 'version', version = f"%(prog)s {__version__}")
        parser.add_argument('path', type = _filter_dir, nargs = '+', help = "target directory")
        parser.add_argument('-v', '--verbose', action = 'count', default = 0, help = "increase verbosity")
        parser.add_argument('-q', '--quiet', action = 'count', default = 0, help = "decrease verbosity")
        parser.add_argument('-m', '--manifest', type = Path, default = Gwak.MANIFEST, metavar = 'FILE', help = f"manifest file (default: {Gwak.MANIFEST})")
        parser.add_argument('--format', choices = libgwak.manifest.formats, default = libgwak.manifest.formats[0], help = "manifest format")
        parser.add_argument('-g', '--grave', type = Path, default = Gwak.GWAK, metavar = 'DIR', help = f"place to bury filebodies (default: {Gwak.GWAK} in first target directory)")
        parser.add_argument('-f', '--force', action = 'store_true', help = "gwak rare or small files, and delete filebodies")
        parser.add_argument('--exclude', type = str, nargs = '*', default = [], metavar = 'DIR', help = "exclude subdirectories by name")
        parser.add_argument('--filter', type = str, default = Gwak.FILTERGLOB, metavar = 'PATTERN', help = f"filter files and subdirectories by glob pattern (default: {Gwak.FILTERGLOB})")
        parser.add_argument('--minsize', type = int, default = Gwak.minsize, metavar = 'N', help = f"minimum file size to be replaced (default: {Gwak.minsize})")
        parser.add_argument('--mindupe', type = int, default = Gwak.mindupe, metavar = 'N', help = f"minimum file appearances to be replaced (default: {Gwak.mindupe})")
        parser.add_argument('--log', type = Path, metavar = 'FILE', help = "tee log to file")
        parser.add_argument('--dry-run', action = 'store_true', help = "do not write anything")
        actions.add_argument('-u', '--undo', '--ungwak', action = 'store_true', help = "ungwak by replacing symlinks with regular files")
        actions.add_argument('--validate', action = 'store_true', help = "validate gwaked directory")
        actions.add_argument('--check', action = 'store_true', help = "integrity check for filebodies")

        __params = parser.parse_args()
        __params.verbosity = __params.verbose - __params.quiet
        __params.isabs = __params.grave.is_absolute()
        if not __params.isabs:
            __params.exclude.append(__params.grave.name)
            __params.grave = __params.path[0] / __params.grave
        __params.manifest = __params.manifest if __params.manifest.is_absolute() else __params.grave / __params.manifest

        __params.logger = logging.getLogger('gwak')
        __params.logger.setLevel(logging.root.level - __params.verbosity * 10 - 1)
        if __params.log:
            __params.logger.addHandler(logf := logging.FileHandler(__params.log.resolve()))
            logf.setFormatter(ZYFormatter(__LOGFORMAT))
        __params.logger.addHandler(logh := logging.StreamHandler())
        logh.setFormatter(ZYFormatter(__LOGFORMAT))


        manifest = libgwak.manifest.Manifest(__params)
        gwak = Gwak(__params, gwakdata = manifest)


        if __params.validate:
            return gwak.validate_files(manifest.load())
        if __params.check:
            return gwak.validate_grave(manifest.load())
        if __params.undo:
            return gwak.redupe(manifest.load())

        manifest.make()
        manifest.write()
        return gwak.dedupe(manifest.get())

    result = run_gwak()
    if __params.verbosity >= 0:
        print(json.dumps(result, indent = 1))

    sys.exit(0 if result else 1)

if __name__ == '__main__':
    main()



import collections
import hashlib
from pathlib import Path

formats = ['yaml', 'json', 'csv']
try:
    import csv
except ImportError:
    formats.remove('csv')
try:
    import json
except ImportError:
    formats.remove('json')
try:
    import yaml
    #yaml.add_representer(collections.defaultdict, yaml.representer.Representer.represent_dict)
except ImportError:
    formats.remove('yaml')

def gwak_size(body: bytes) -> str:
    return '{:016x}'.format(len(body))

def gwak_hash(body: bytes) -> str:
    return hashlib.sha3_512(body).hexdigest()

class Manifest():
    _data: dict = {}
    _file: Path = None
    _logger = None
    _params: dict = {}


    def __init__(self, params: dict):
        self._params = params
        self._logger = params.logger
        self._file = params.manifest

    def _normalize(self, data: dict) -> dict:
        return {k: dict(v) for k, v in data.items()}

    def _walk(self, path: Path):
        for item in path.iterdir():
            if item.name in self._params.exclude or not item.match(self._params.filter):
                self._logger.info(f"excluding path [{item}]")
                continue
            if item.is_symlink():
                self._logger.debug(f"skipping symlink [{item}]")
                continue
            if item.is_dir():
                yield from self._walk(item)
                continue
            if not item.is_file():
                self._logger.debug(f"skipping irregular file [{item}]")
                continue
            yield item.resolve()

    def _gen_bytree(self):
        for path in self._params.path:
            for file in self._walk(path):
                body = file.read_bytes()
                yield gwak_size(body), gwak_hash(body), file.resolve()

    def _gen_bytable(self, files):
        for size, hash, file in files:
            yield size, hash, Path(file)

    def _csvload(self, f) -> dict:
        return self._makedict(self._gen_bytable(csv.reader(f)))

    def _makedict(self, gen) -> dict:
        gwaks = collections.defaultdict(lambda: collections.defaultdict(list))
        for size, hash, file in gen:
            gwaks[size][hash].append(file)
        return self._normalize(gwaks)

    def _backup(self) -> None:
        self._logger.warning(f"backing up old manifest [{self._file}]")
        hash = hashlib.sha3_512(self._file.read_bytes()).hexdigest()
        return self._file.rename(f"{self._file}.{hash}")

    def _read(self) -> dict:
        with self._file.open(mode = 'r') as f:
            match self._params.format:
                case 'yaml':
                    return yaml.safe_load(f)
                case 'json':
                    return json.load(f)
                case 'csv':
                    return self._csvload(f)

    def _write(self) -> None:
        with self._file.open(mode = 'w') as f:
            match self._params.format:
                case 'yaml':
                    return yaml.dump(self.serialize(), f)
                case 'json':
                    return json.dump(self.serialize(), f)
                case 'csv':
                    return csv.writer(f).writerows(self.pettan())


    def get(self) -> dict:
        return self._data

    def read(self) -> dict:
        self._logger.info(f"reading manifest [{self._file}]")
        if self._params.format not in formats:
            raise NotImplementedError(self._params.format)
        return self._read()

    def load(self) -> dict:
        gwaks = self.read()
        for size, gwak in gwaks.items():
            for hash, files in gwak.items():
                files[:] = (Path(file) for file in files)
        self._data = gwaks
        return self._data

    def write(self) -> None | bool:
        self._logger.info(f"writing manifest [{self._file}]")
        if self._params.dry_run:
            return True
        self._file.parent.mkdir(parents = True, exist_ok = True)
        if self._file.is_file():
            self._backup()
        if self._params.format not in formats:
            raise NotImplementedError(self._params.format)
        return self._write()

    def make(self) -> dict:
        self._data = self._makedict(self._gen_bytree())
        return self._data

    def pettan(self) -> list:
        return [(s,h,str(p))for s,g in self._data.items()for h,f in g.items()for p in f]

    def serialize(self) -> dict:
        return {s:{h:[str(p)for p in f]for h,f in g.items()}for s,g in self._data.items()}



import setuptools

setuptools.setup(
    name = "gwak",
    version = "0.5.1",
    author = "LiviaMedeiros",
    author_email = "livia@cirno.name",
    description = "Directory gwaking utility",
    long_description = "Gwak a directory by burying filebodies and replacing them with symlinks.",
    long_description_content_type = "text/plain",
    license = "GPLv3",
    url = "https://github.com/LiviaMedeiros/gwak",
    project_urls = {
        "Bug Tracker": "https://github.com/LiviaMedeiros/gwak/issues",
    },
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    py_modules = [
        "gwak",
        "libgwak.manifest",
        "libgwak.zy"
    ],
    package_dir = {
        '': 'src'
    },
    entry_points = {
        'console_scripts': [
            'gwak = gwak:main'
        ]
    },
    python_requires = ">=3.10",
)


#!/usr/bin/env python

import datetime
import re

class ZY:
    r: re = re.compile(r"^([\d]+)-([\d]+)-([\d]+)_([\d]+):([\d]+):([\d]+)$")

    def __init__(self, base: str = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwx', rx: str = '[0-9A-Za-x]'):
        self.z = tuple(base)
        self.b = {i:s for s,i in enumerate(base)}
        self.c = len(base)
        self.rx = rx

    def _enc(self, n: int) -> str:
        a = ''
        while True:
            a = self.z[n % self.c] + a
            n //= self.c
            if not n:
                return a

    def _dec(self, d: str) -> int:
        t = 0
        m = 1
        for n in range(len(s := tuple(d)) - 1, -1, -1):
            t += self.b[s[n]] * m
            m *= self.c
        return t

    def encode(self, time: datetime = None) -> str:
        if time is None:
            time = datetime.datetime.utcnow()
        return ''.join(map(self._enc,map(int,m.groups())))if(m:=self.r.match(time.strftime('%Y-%m-%d_%H:%M:%S')))else False

    def decode(self, zytime: str, oformat: str = '%04d-%02d-%02dT%02d:%02d:%02dZ', iformat: str = '(__)(_)(_)(_)(_)(_)') -> str:
        return oformat%tuple(map(self._dec,m.groups()))if(m:=re.match(fr"^z*{iformat}y*$".replace('_',self.rx),zytime))else False

zy = ZY()

def main():
    import argparse
    parser = argparse.ArgumentParser()
    actions = parser.add_mutually_exclusive_group()
    actions.add_argument('-e', '--encode', type = datetime.datetime.fromisoformat, metavar = 'ISO_time')
    actions.add_argument('-d', '--decode', type = str, metavar = 'ZY_time')
    __params = parser.parse_args()

    if __params.encode:
        return zy.encode(__params.encode)
    if __params.decode:
        return zy.decode(__params.decode)
    return zy.encode()

if __name__ == '__main__':
    print(main())
